package kc.mega.dodge;

import kc.mega.game.BattleField;
import kc.mega.game.BotState;
import kc.mega.game.GameState;
import kc.mega.game.Predict;
import kc.mega.game.PredictState;
import kc.mega.misc.Strategy;
import kc.mega.utils.Geom;
import kc.mega.utils.KUtils;
import kc.mega.utils.Painter;
import kc.mega.wave.Wave;

import java.awt.Color;
import java.awt.geom.Point2D;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import jk.math.FastTrig;
import robocode.AdvancedRobot;

public class PathSurfer {
  private static final boolean TRUE_SURF = false;
  private static final double TARGET_DISTANCE = 600;

  private final AdvancedRobot bot;
  private final Strategy strategy;
  private final DangerCalculator dangerComputer;
  private final Paths paths;

  private GameState gs;
  private List<MovementWave> waves;
  private List<Integer> currentPath;
  private List<SurfPlan> plans;
  private SurfPlan emptyPlan;

  public PathSurfer(AdvancedRobot bot, Strategy strategy, DangerCalculator dangerComputer) {
    this.bot = bot;
    this.strategy = strategy;
    this.dangerComputer = dangerComputer;
    this.paths = new Paths();
  }

  public void surf(GameState gs, List<MovementWave> waves, boolean paint) {
    this.gs = gs;
    this.waves = waves;
    BotState myState = gs.getMyState(), enemyState = gs.getEnemyState();
    double distance = gs.getMyState().distance(enemyState);

    plans = new ArrayList<>();
    emptyPlan = new SurfPlan(myState.asPredictState());
    plans.add(emptyPlan);

    if (waves.size() > 0) {
      if (TRUE_SURF || strategy.antiRam) {
        trueSurfBranch();
      } else {
        pathSurfBranch();
        if (currentPath != null && currentPath.size() > 1) {
          currentPath.remove(0);
          addIfNotDuplicate(currentPath, plans, emptyPlan.getSurfWave());
        }
      }
      setDangerMultipliers();
      if (TRUE_SURF) {
        trueSurfBranch();
      } else {
        setNextWaveDangers();
      }
    } else {
      noWaveBranch();
      setDangerMultipliers();
    }

    List<SurfPlan> rankedPlans = getRankedPlans();
    SurfPlan bestPlan = rankedPlans.get(0);
    currentPath = bestPlan.path;
    List<MoveInstruction> instructions = bestPlan.getInstructions();

    BotState futureState = gs.getMyState();
    List<BotState> futureStates = new ArrayList<>();
    for (int i = 0; i < instructions.size(); i++) {
      MoveInstruction instruction = instructions.get(i);
      futureState = futureState.getNextState(instruction.direction, instruction.turn);
      futureStates.add(futureState);
    }
    if (paint) {
      SurfPlan fullPlan = bestPlan.nextWavePlan == null ? bestPlan : bestPlan.nextWavePlan;
      for (int i = 1; i < fullPlan.states.size(); i++) {
        Painter.THIS_TICK.addPoint(fullPlan.path.get(i - 1) == 0 ? Color.gray.brighter() :
          (fullPlan.path.get(i - 1) == 1 ? Color.yellow : Color.red),
          fullPlan.states.get(i).location, 2);
        if (i == bestPlan.states.size() - 1) {
          Painter.THIS_TICK.addBot(Color.white, fullPlan.states.get(i).location);
        }
        if (bestPlan.nextWavePlan != null && i == bestPlan.nextWavePlan.states.size() - 1) {
          Painter.THIS_TICK.addBot(Color.gray, fullPlan.states.get(i).location);
        }
      }

      if (waves.size() > 0 && bestPlan.visitOffsetRanges.containsKey(waves.get(0))) {
        paintVisitOffsets(waves.get(0), bestPlan, Color.white);
        if (bestPlan.nextWavePlan != null) {
          paintVisitOffsets(waves.get(1), bestPlan.nextWavePlan, Color.gray);
        }
      }
      if (bestPlan.approximateWaveLoc != null) {
        MovementWave wave = waves.get(2);
        wave.paintTick(Color.gray, wave.getOffset(
            dangerComputer.getBins().midPoint[bestPlan.approximateWaveLoc.start]));
        wave.paintTick(Color.gray, wave.getOffset(
            dangerComputer.getBins().midPoint[bestPlan.approximateWaveLoc.end]));
      }
    }
    gs.setMyNextStates(futureStates);

    bot.setTurnRightRadians(instructions.get(0).turn);
    int direction = instructions.get(0).direction;
    bot.setAhead(direction == 0 ? (Math.random() - 0.5) / 1e12 :
      direction * Double.POSITIVE_INFINITY);
    if (strategy.ram) {
      double targetHeading = getTargetHeading(myState, bestPlan.path.get(0),
          myState.absoluteBearing(enemyState), distance, 0);
      double turn = FastTrig.normalRelativeAngle(targetHeading - bot.getHeadingRadians());
      bot.setMaxVelocity(Math.max(0, Math.abs(turn) < Math.PI / 9 ? 8 :
        8 - Math.abs(turn / (Math.PI / 6))));
    } else {
      bot.setMaxVelocity(8 - Math.random() / 1e12);
    }
  }

  public void paintVisitOffsets(MovementWave surfWave, SurfPlan plan, Color color) {
    if (surfWave.radius > 0 && plan.visitOffsetRanges.containsKey(surfWave)) {
      double[] offsets = plan.visitOffsetRanges.get(surfWave);
      surfWave.paintTick(color, offsets[0]);
      surfWave.paintTick(color, offsets[1]);
    }
  }

  public void addTrueSurfChildren(SurfPlan parent, List<SurfPlan> surfPlans, int ticks) {
    MovementWave surfWave = parent.getSurfWave();
    if (strategy.antiRam && parent == emptyPlan) {
      ticks = Math.max(ticks, 8);
    }
    SurfPlan[] children = strategy.antiRam ?
        new SurfPlan[] {
            new SurfPlan(parent, Paths.FORWARD_PATH, 1),
            new SurfPlan(parent, Paths.BACKWARD_PATH, 1),
            new SurfPlan(parent, Paths.HALT_PATH, 1),
            new SurfPlan(parent, Paths.FORWARD_PATH, -1),
            new SurfPlan(parent, Paths.BACKWARD_PATH, -1),
            new SurfPlan(parent, Paths.HALT_PATH, -1)} :
         new SurfPlan[] {
            new SurfPlan(parent, Paths.FORWARD_PATH),
            new SurfPlan(parent, Paths.BACKWARD_PATH),
            new SurfPlan(parent, Paths.HALT_PATH)};
    for (SurfPlan child : children) {
      child.predict(surfWave, ticks, true);
      surfPlans.add(child);
    }
  }

  public void noWaveBranch() {
    plans.clear();
    addTrueSurfChildren(emptyPlan, plans, 8);
  }

  public void trueSurfBranch() {
    List<SurfPlan> newPlans = new ArrayList<>();
    for (SurfPlan parent : plans) {
      if (parent.getSurfWave() == null) {
        newPlans.add(parent);
      } else {
        addTrueSurfChildren(parent, newPlans, -1);
      }
    }
    plans = newPlans;
  }

  public void pathSurfBranch() {
    List<SurfPlan> newPlans = new ArrayList<>();
    SurfPlan forward = new SurfPlan(emptyPlan, Paths.FORWARD_PATH);
    SurfPlan backward = new SurfPlan(emptyPlan, Paths.BACKWARD_PATH);
    MovementWave surfWave = emptyPlan.getSurfWave();
    forward.predict(surfWave);
    backward.predict(surfWave);
    newPlans.add(forward);
    newPlans.add(backward);

    CandidatePath forwardCandidate = getCandidatePath(surfWave, forward);
    CandidatePath backwardCandidate = getCandidatePath(surfWave, backward);
    if (forwardCandidate != null) {
      addIfNotDuplicate(forwardCandidate.path, newPlans, surfWave);
    }
    if (backwardCandidate != null) {
      addIfNotDuplicate(backwardCandidate.path, newPlans, surfWave);
    }
    addIfNotDuplicate(Paths.HALT_PATH, newPlans, surfWave);
    plans = newPlans;
  }

  private static class CandidatePath {
    public final List<Integer> path;
    public final double danger;

    public CandidatePath(List<Integer> path, double danger) {
      this.path = path;
      this.danger = danger;
    }
  }

  public CandidatePath getCandidatePath(MovementWave surfWave, SurfPlan plan) {
    return getCandidatePath(surfWave, plan,
        Math.random() < 0.5 ? 0 : (Math.random() < 0.5 ? -8 : 8));
  }

  public CandidatePath getCandidatePath(MovementWave surfWave, SurfPlan plan, double targetVel) {
    if (plan.states.size() <= 2) {
      return null;
    }
    Point2D.Double bestPoint = null;
    double bestDanger = Double.MAX_VALUE;
    double bestDistance = 0;
    for (int i = 0; i < plan.states.size(); i++) {
      double danger = dangerComputer.getApproximatePointDanger(
          surfWave, plan.states.get(i).location);
      if (danger < bestDanger) {
        bestDanger = danger;
        bestDistance = plan.travelDistances.get(i);
        bestPoint = plan.states.get(i).location;
      }
    }
    List<Integer> path = paths.bestPath(plan.states.get(0).velocity,
        Math.max(1, surfWave.getTicksUntilPasses(bestPoint, gs.gameTime) - 1),
        bestDistance, targetVel);
    return new CandidatePath(path, bestDanger);
  }

  public void addIfNotDuplicate(List<Integer> path, List<SurfPlan> plans, MovementWave surfWave) {
    for (SurfPlan plan : plans) {
      int pathDirection = 0;
      int planDirection = 0;
      boolean different = false;
      for (int i = 0; i < Math.max(path.size(), plan.path.size()); i++) {
        if (i < path.size()) {
          pathDirection = path.get(i);
        }
        if (i < plan.path.size()) {
          planDirection = plan.path.get(i);
        }
        if (pathDirection != planDirection) {
          different = true;
          break;
        }
      }
      if (!different) {
        return;
      }
    }
    SurfPlan plan = new SurfPlan(emptyPlan, path);
    plan.predict(surfWave);
    plans.add(plan);
  }

  public void setDangerMultipliers() {
    for (SurfPlan plan : plans) {
      plan.setDangerMultiplier();
    }
  }

  public void setNextWaveDangers() {
    double bestDanger = Double.POSITIVE_INFINITY;
    for (SurfPlan plan : getRankedPlans()) {
      if (plan.getDanger() > bestDanger) {
        break;
      }
      plan.setNextWaveDanger(bestDanger);
      double danger = plan.getDanger();
      if (danger < bestDanger) {
        bestDanger = danger;
      }
    }
  }

  public List<SurfPlan> getRankedPlans() {
    for (SurfPlan plan : plans) {
      plan.setWaveDanger();
    }
    Collections.sort(plans);
    return plans;
  }

  public static double getWaveWeight(MovementWave w) {
    return (0.5 + w.power) / Math.sqrt(2 + Math.max(1, w.ticksUntilBreak));
  }

  private class SurfPlan implements Comparable<SurfPlan> {
    public final List<PredictState> states;
    public final Map<MovementWave, double[]> visitOffsetRanges;
    public final List<Double> travelDistances;
    public final int antiRamEvadeSide;
    public List<Integer> path;
    public boolean hitWall;
    private PredictState enemyState;

    public double waveDanger;
    public double nextWaveDanger;
    public double dangerMultiplier;
    public SurfPlan nextWavePlan;
    public DangerCalculator.ApproximateDangerLocation approximateWaveLoc;

    public SurfPlan(PredictState state) {
      states = new ArrayList<>();
      states.add(state);
      travelDistances = new ArrayList<>();
      travelDistances.add(0.0);
      path = new ArrayList<>();
      visitOffsetRanges = new HashMap<>();
      hitWall = false;
      waveDanger = nextWaveDanger = 0;
      dangerMultiplier = 1;
      antiRamEvadeSide = 1;
    }

    public SurfPlan(SurfPlan parent, List<Integer> extension) {
      this(parent, extension, 1);
    }

    public SurfPlan(SurfPlan parent, List<Integer> extension, int antiRamEvadeSide) {
      this.antiRamEvadeSide = antiRamEvadeSide;
      states = new ArrayList<>(parent.states);
      travelDistances = new ArrayList<>(parent.travelDistances);
      path = new ArrayList<>(parent.path);
      visitOffsetRanges = new HashMap<>();
      for (Map.Entry<MovementWave, double[]> e : parent.visitOffsetRanges.entrySet()) {
        this.visitOffsetRanges.put(e.getKey(), e.getValue().clone());
      }
      hitWall = parent.hitWall;
      waveDanger = nextWaveDanger = 0;
      dangerMultiplier = parent.dangerMultiplier;
      path = new ArrayList<>(parent.path);
      path.addAll(extension);
    }

    private void predict(MovementWave surfWave) {
      predict(surfWave, -1, true);
    }

    private void predict(MovementWave surfWave, int minTicks, boolean precise) {
      PredictState state = getEndState();
      int tick = 0;
      int direction = 0;
      int nonzeroDirection = 0;

      enemyState = gs.getEnemyState();
      boolean predictEnemy = strategy.antiRam && states.size() == 1;
      boolean ICollided = gs.lastCollideTime == gs.gameTime;
      boolean enemyCollided = false;

      double distanceTraveled = travelDistances.get(travelDistances.size() - 1);
      boolean surfWavePassed = surfWave == null;
      while (!surfWavePassed || tick <= minTicks) {
        //Timer.INSTANCE.start("Next Tick Predict");
        tick = (int)(state.gameTime - states.get(0).gameTime);
        if (tick < path.size()) {
          direction = path.get(tick);
          if (direction != 0) {
            nonzeroDirection = direction;
          }
        } else {
          path.add(direction);
        }

        if (predictEnemy) {
          double ramHeading = enemyState.absoluteBearing(state);
          ramHeading += state.velocity * Math.sin(state.heading - ramHeading) / 15;
          double enemyTurn = FastTrig.normalRelativeAngle(ramHeading - enemyState.heading);
          int enemyDirection = 1;
          if (enemyTurn < -Math.PI / 2) {
            enemyTurn += Math.PI;
            enemyDirection = -1;
          } else if (enemyTurn > Math.PI / 2) {
            enemyTurn -= Math.PI;
            enemyDirection = -1;
          }
          PredictState nextEnemyState = enemyState.getNextState(
              enemyDirection, enemyCollided ? 0 : enemyTurn);
          nextEnemyState = BattleField.INSTANCE.wallCollisionCheck(nextEnemyState, true);
          if (enemyCollided = nextEnemyState.collidesWith(state)) {
            enemyState = new PredictState(
                enemyState.location, nextEnemyState.heading, 0, nextEnemyState.gameTime);
          } else {
            enemyState = nextEnemyState;
          }
        }

        double absoluteBearing = state.absoluteBearing(gs.getEnemyState());
        double distance = state.distance(gs.getEnemyState());
        double targetHeading = getTargetHeading(
            state, nonzeroDirection == 0 ? KUtils.nonzeroSign(state.velocity) : nonzeroDirection,
                absoluteBearing, distance, antiRamEvadeSide);
        double turn = ICollided ? 0 : FastTrig.normalRelativeAngle(targetHeading - state.heading);
        double prevVel = state.velocity;
        state = state.getNextState(direction, turn);
        PredictState inFieldState = BattleField.INSTANCE.wallCollisionCheck(state, true);
        if (inFieldState != state) {
          state = inFieldState;
          if (Math.abs(prevVel) > 2) {
            hitWall = true;
          }
        }
        if (ICollided = (predictEnemy && state.collidesWith(enemyState))) {
          state = new PredictState(getEndState().location, state.heading, 0, state.gameTime);
        }
        states.add(state);
        distanceTraveled += state.velocity;
        travelDistances.add(distanceTraveled);
        //Timer.INSTANCE.stop();

        //Timer.INSTANCE.start("Handle Waves");
        //TODO: ignore far-off waves, make this faster
        if (precise) {
          for (MovementWave w : waves) {
            Wave.Foam foam = w.getFoam(state);
            if (w == surfWave && foam.status >= Wave.WILL_PASS_THIS_TICK) {
              surfWavePassed = true;
            }
            if (foam.status == Wave.BREAKING || foam.status == Wave.WILL_PASS_THIS_TICK) {
              double[] offsetRange = visitOffsetRanges.get(w);
              if (offsetRange == null) {
                visitOffsetRanges.put(w, foam.hitOffsetRange);
              } else {
                Geom.updateOffsetRange(offsetRange, foam.hitOffsetRange);
              }
            }
          }
        } else {
          surfWavePassed = surfWave.source.distance(state.location) <
              (1 + state.gameTime - surfWave.fireTime) * surfWave.speed;
        }
        //Timer.INSTANCE.stop();
      }
    }

    public MovementWave getSurfWave() {
      PredictState state = getEndState();
      int minTicksUntilPasses = 1000;
      MovementWave surfWave = null;
      for (MovementWave w : waves) {
        int ticksUntilPasses = w.getTicksUntilPasses(state);
        if (ticksUntilPasses > 1 && ticksUntilPasses < minTicksUntilPasses) {
          minTicksUntilPasses = ticksUntilPasses;
          surfWave = w;
        }
      }
      return surfWave;
    }

    public PredictState getEndState() {
      return states.get(states.size() - 1);
    }

    public void setWaveDanger() {
      waveDanger = 0;
      for (Map.Entry<MovementWave, double[]> e : visitOffsetRanges.entrySet()) {
        waveDanger += dangerComputer.getDanger(e.getKey(), e.getValue()) *
            getWaveWeight(e.getKey());
      }
    }

    public void setNextWaveDanger(double minDanger) {
      MovementWave surfWave = getSurfWave();
      if (surfWave == null) {
        nextWaveDanger = 0;
        return;
      }
      SurfPlan forward = makeNextWavePlan(surfWave, Paths.FORWARD_PATH);
      SurfPlan backward = makeNextWavePlan(surfWave, Paths.BACKWARD_PATH);
      SurfPlan halt = makeNextWavePlan(surfWave, Paths.HALT_PATH);
      CandidatePath forwardCandidate = getCandidatePath(surfWave, forward, 0);
      CandidatePath backwardCandidate = getCandidatePath(surfWave, backward, 0);
      CandidatePath bestCandidate;
      if (forwardCandidate != null && backwardCandidate != null) {
        bestCandidate = forwardCandidate.danger < backwardCandidate.danger ?
            forwardCandidate : backwardCandidate;
      } else if (forwardCandidate == null) {
        bestCandidate = backwardCandidate;
      } else {
        bestCandidate = forwardCandidate;
      }
      nextWavePlan = forward.waveDanger < backward.waveDanger ? forward : backward;
      nextWavePlan = halt.waveDanger < nextWavePlan.waveDanger ? halt : nextWavePlan;
      if (bestCandidate != null) {
        SurfPlan plan = makeNextWavePlan(surfWave, bestCandidate.path);
        nextWavePlan = plan.waveDanger < nextWavePlan.waveDanger ? plan : nextWavePlan;
      }

      nextWaveDanger = nextWavePlan.waveDanger;
      if (getDanger() < minDanger) {
        nextWaveDanger += 0.75 * nextWavePlan.getApproximateDanger();
        approximateWaveLoc = nextWavePlan.approximateWaveLoc;
      }
    }

    public double getApproximateDanger() {
      MovementWave surfWave = getSurfWave();
      if (surfWave == null || surfWave.isVirtual) {
        return 0;
      }
      SurfPlan forward = new SurfPlan(this, Paths.FORWARD_PATH);
      SurfPlan backward = new SurfPlan(this, Paths.BACKWARD_PATH);
      forward.predict(surfWave, -1, false);
      backward.predict(surfWave, -1, false);
      approximateWaveLoc = dangerComputer.getBestApproximateDanger(
          surfWave, Math.min(getEndState().location.distance(surfWave.source),
                    Math.min(forward.getEndState().location.distance(surfWave.source),
                             backward.getEndState().location.distance(surfWave.source))),
          Geom.offset(surfWave.source, forward.getEndState().location,
              surfWave.absoluteBearing),
          Geom.offset(surfWave.source, backward.getEndState().location,
              surfWave.absoluteBearing));
      return getWaveWeight(surfWave) * approximateWaveLoc.danger;
    }

    public SurfPlan makeNextWavePlan(MovementWave surfWave, List<Integer> path) {
      SurfPlan plan = new SurfPlan(this, path);
      plan.visitOffsetRanges.clear();
      plan.predict(surfWave);
      plan.setWaveDanger();
      return plan;
    }

    public void setDangerMultiplier() {
      Point2D.Double endLocation = getEndState().location;
      Point2D.Double enemyLocation = gs.getEnemyState().location;
      double distanceRatio = states.get(0).location.distance(enemyLocation) /
          endLocation.distance(enemyLocation);
      dangerMultiplier = Math.exp(distanceRatio * (strategy.ram ? -1.4 : 1.4));
      if (strategy.antiRam) {
        dangerMultiplier /= Math.sqrt(18 + BattleField.INSTANCE.wallDistance(endLocation));
        dangerMultiplier /= getEndState().distance(enemyState);
        dangerMultiplier /= Math.pow(Geom.wallRestriction(enemyLocation, endLocation, 9), 1.5);
      } else {
        double bulletSpeed = KUtils.bulletSpeed(gs.lastEnemyBulletPower);
        dangerMultiplier *= Math.pow(Geom.wallRestriction(
            endLocation, enemyLocation, bulletSpeed), 1);
        dangerMultiplier /= Math.pow(Geom.wallRestriction(
            enemyLocation, endLocation, bulletSpeed), 2);
      }
    }

    public double getDanger() {
      return (
          0.01 + waveDanger + nextWaveDanger +
          strategy.getAntiMirrorDanger(getEndState(), waveDanger + nextWaveDanger) +
          (hitWall ? 0.2 : 0) +
          ((strategy.antiRam || strategy.ram) ? 0 :
            Math.max(0, 1 - getEndState().distance(gs.getEnemyState()) / 300))
          ) * dangerMultiplier;
    }

    public List<MoveInstruction> getInstructions() {
      List<MoveInstruction> instructions = new ArrayList<>();
      for (int i = 0; i < states.size() - 1; i++) {
        instructions.add(new MoveInstruction(path.get(i),
            FastTrig.normalRelativeAngle(states.get(i + 1).heading - states.get(i).heading)));
      }
      return instructions;
    }

    @Override
    public int compareTo(SurfPlan other){
      return (int)Math.signum(getDanger() - other.getDanger());
    }
  }

  public static class MoveInstruction {
    public final int direction;
    public final double turn;

    public MoveInstruction(int direction, double turn) {
      this.direction = direction;
      this.turn = turn;
    }
  }

  public double getTargetHeading(
      PredictState state,
      int direction,
      double absoluteBearing,
      double distance,
      int antiRamEvadeSide) {
    //Timer.INSTANCE.start("Heading");
    absoluteBearing += Math.PI;
    double nextVelocity = Predict.getNextVelocity(state.velocity, direction);
    direction = nextVelocity == 0 ? direction : (int)Math.signum(nextVelocity);
    int orbitDirection = KUtils.orbitDirection(absoluteBearing, state.heading, direction);

    double targetHeading;
    if (strategy.antiRam) {
      targetHeading = absoluteBearing + antiRamEvadeSide * Math.max(0.7, 1.5 - distance / 400);
      if (direction == -1) {
        targetHeading += Math.PI;
      }
    } else if (strategy.ram) {
      PredictState enemyFutureState = gs.getEnemyState();
      for (int i = 0; i < Math.min(10, distance / 10); i++) {
        enemyFutureState = enemyFutureState.getNextState(
            (int)Math.signum(gs.getEnemyState().velocity),
            FastTrig.normalRelativeAngle(gs.getEnemyState().heading -
                gs.enemyHistory.get(1).heading));
      }
      targetHeading = state.absoluteBearing(enemyFutureState);
      if (direction == -1) {
        targetHeading += Math.PI;
      }
      return targetHeading;
    } else {
      //double attackAngle = (distance - TARGET_DISTANCE) / (waves.size() == 0 ?  400 : 600);
      double attackAngle = -Math.min(0.9, (waves.size() == 0 ? 1.5 : 1) *
          KUtils.sqr(Math.max(650 - distance, 0)) / 220000);
      direction = nextVelocity == 0 ? direction : (int)Math.signum(nextVelocity);
      targetHeading = absoluteBearing + orbitDirection * (attackAngle + direction * Math.PI / 2);
    }

    double turn = FastTrig.normalRelativeAngle(targetHeading - state.heading);
    double nextHeading = state.heading + Predict.getTurnIncrement(
        turn, state.velocity);
    if (strategy.walkingStickSmooth) {
      // staying away from walls makes simple targeters with wall collision checks easier to predict
      targetHeading = BattleField.INSTANCE.walkingStickSmooth(
          state.location, targetHeading, direction, orbitDirection, 150, 25);
      turn = FastTrig.normalRelativeAngle(targetHeading - state.heading);
      nextHeading = state.heading + Predict.getTurnIncrement(turn, state.velocity);
    }
    double offset = direction == 1 ? 0 : Math.PI;
    return BattleField.INSTANCE.fancyStickSmooth(
        FastTrig.normalAbsoluteAngle(nextHeading + offset),
        Math.abs(nextVelocity), state.location.x, state.location.y,
        orbitDirection) + offset;
  }
}
